(*
 *  ݵרõԪ
 *
 *  C/SWebSocket ģʽݽ
 *)
unit iocp_receivers;

interface

{$I in_iocp.inc}        // ģʽ

uses
  {$IFDEF DELPHI_XE7UP}
  Winapi.Windows, System.Classes, System.SysUtils, {$ELSE}
  Windows, Classes, SysUtils, {$ENDIF}
  iocp_base, iocp_md5, iocp_mmHash,
  iocp_msgPacks, iocp_WsJSON;

type


  // ============ ݽ  =============

  TBaseReceiver = class(TObject)
  protected
    FBuffer: PAnsiChar;         // ڴ
    FBufSize: Cardinal;         // δݳ
    FRecvCount: TFileSize;      // ÿϢյֽ
    FCancel: Boolean;           // ȡWS
    FCheckPassed: Boolean;      // УɹWS
    FComplete: Boolean;         // ȫ
    FErrorCode: Integer;        // 쳣
    procedure IncBufferPos(Offset: Cardinal); {$IFDEF USE_INLINE} inline; {$ENDIF}
    procedure InterInit(ACancel, AComplete: Boolean); {$IFDEF USE_INLINE} inline; {$ENDIF}
    procedure SetComplete(const Value: Boolean); 
  public
    constructor Create;
    procedure Clear; virtual;
    procedure Prepare(const AData: PAnsiChar; const ASize: Cardinal); virtual; abstract;
    procedure Receive(const AData: PAnsiChar; const ASize: Cardinal); virtual; abstract;
    procedure Reset; virtual;   //  WS
  public
    property CheckPassed: Boolean read FCheckPassed;  
    property Complete: Boolean read FComplete write SetComplete;
    property ErrorCode: Integer read FErrorCode;    
  end;

  // ============ C/S ģʽݽ  =============

  TCSBaseReceiver = class(TBaseReceiver)
  private
    FOwner: TReceivePack;       // Ϣ
    FCheckCode: TIOCPHashCode;  // У
    FCheckCode2: TIOCPHashCode; // У
    FMainLackSize: Cardinal;    // Ƿȱݳ
    FAttachLackSize: TFilesize; // Ƿȱݳ
    FReadHead: Boolean;         // ѾȡЭͷϢ
    procedure ExtractMessage;
    procedure VerifyMainStream;
    procedure VerifyAttachmentStream;
    procedure UZipPretreatMainStream;
    procedure UZipTreatAttachmentStream;
  protected
    procedure IncRecvCount(RecvCount, DataType: Cardinal); virtual; 
    procedure RecvSubqueueData; virtual;
    procedure ReceiveMainFinish; virtual;
    procedure ReceiveAttachmentFinish; virtual;
    procedure WriteMainStream(ByteCount: Cardinal); virtual;
    procedure WriteAttachmentStream(ByteCount: Cardinal); virtual;
  protected
    procedure CreateAttachment; virtual; abstract;
    procedure GetCheckCodes; virtual; abstract;
  public
    procedure Clear; override;
    procedure Reset; override;
    procedure OwnerClear;
  public
    property Cancel: Boolean read FCancel;
    property Owner: TReceivePack read FOwner;
  end;

  // ============ ݽ  =============
  // 򸽼ݽ -> Ӧò

  TServerReceiver = class(TCSBaseReceiver)
  protected
    procedure IncRecvCount(RecvCount, DataType: Cardinal); override;
  protected
    procedure CreateAttachment; override;
    procedure GetCheckCodes; override;
  public
    constructor Create(AOwner: TReceivePack);
    procedure Prepare(const AData: PAnsiChar; const ASize: Cardinal); override;
    procedure Receive(const AData: PAnsiChar; const ASize: Cardinal); override;
  end;

  // ============ ͻݽ  =============
  // ͬʱյϢϢһϢͶŵӦò㣬
  // ݰϢʱЭͷУ뱻۶ϡ

  // Эͷռ
  TMsgHeadBuffers   = array[0..IOCP_SOCKET_SIZE - 1] of AnsiChar;

  // У쳣¼
  TCheckErrorEvent  = procedure(Result: TReceivePack) of object;

  // ύӦò¼
  TPostMessageEvent = procedure(Result: TReceivePack) of object;

  // Ϣ¼
  TReceiveDataEvent = procedure(Result: TReceivePack; DataType: TMessageDataType;
                                ReceiveCount: TFileSize; AttachFinished: Boolean) of object;

  TClientReceiver = class(TCSBaseReceiver)
  private
    // Ϣ
    FOwnerClass: TReceivePackClass;
    
    // ۶ϵЭͷ
    FHeadBuffers: TMsgHeadBuffers;
    FHeadLackSize: Cardinal;  // Эȱݳ

    FHashCode: PAnsiChar;     // дУַ
    FCodeLackSize: Cardinal;  // Уȱٳ
    FLocalPath: string;       // ·

    FOnError: TCheckErrorEvent;    // У쳣¼
    FOnPost: TPostMessageEvent;    // ͶϢ
    FOnReceive: TReceiveDataEvent; // Ϣ¼

    procedure ScanRecvBuffers;
    procedure WriteHashCode(ByteCount: Cardinal);
  protected
    procedure IncRecvCount(RecvCount, DataType: Cardinal); override;
    procedure ReceiveMainFinish; override;
    procedure ReceiveAttachmentFinish; override;
    procedure WriteAttachmentStream(ByteCount: Cardinal); override;
  protected
    procedure CreateAttachment; override;
    procedure GetCheckCodes; override;
  public
    constructor Create(OwnerClass: TReceivePackClass);
    destructor Destroy; override;
    procedure PostMessage;
    procedure Prepare(const AData: PAnsiChar; const ASize: Cardinal); override;
    procedure Receive(const AData: PAnsiChar; const ASize: Cardinal); override;
  public
    property Complete: Boolean read FComplete;
    property LocalPath: String read FLocalPath write FLocalPath;
  public
    property OnError: TCheckErrorEvent read FOnError write FOnError;
    property OnPost: TPostMessageEvent read FOnPost write FOnPost;
    property OnReceive: TReceiveDataEvent read FOnReceive write FOnReceive;
  end;

  // ============ WebSocket ݽ =============

{ byte: 0               1               2               3
   bit: 7 6 5 4 3 2 1 0 7 6 5 4 3 2 1 0 7 6 5 4 3 2 1 0 7 6 5 4 3 2 1 0
       +-+-+-+-+-------+-+-------------+-------------------------------+
       |F|R|R|R| opcode|M| Payload len |    Extended payload length    |
       |I|S|S|S|  (4)  |A|     (7)     |             (16/64)           |
       |N|V|V|V|       |S|             |   (if payload len==126/127)   |
       | |1|2|3|       |K|             |                               |
       +-+-+-+-+-------+-+-------------+ - - - - - - - - - - - - - - - +
       |     Extended payload length continued, if payload len == 127  |
       + - - - - - - - - - - - - - - - +-------------------------------+
       |                               |Masking-key, if MASK set to 1  |
       +-------------------------------+-------------------------------+
       | Masking-key (continued)       |          Payload Data...      |
       +-------------------------------- - - - - - - - - - - - - - - - + }

  // עWebSocket ֶ֧֡֡Ѿִ֧ݴ䣬
  //    ֡ݣÿ֡СʱӽĸԣȽá
  // һж֡ʱֻյ֡ͻȫմ

  TWSBaseReceiver = class(TBaseReceiver)
  private
    FOwner: TObject;           // 
    
    FHeader: TWebSocketFrame;  // ۶ϵЭͷ
    FHeadAddr: PAnsiChar;      // д FFrameHeader λ
    FLackSize: Cardinal;       // ֡ݲĳ

    FData: PAnsiChar;          // ݿʼλ
    FFrameSize: UInt64;        // ǰ֡ݴС
    FFrameRecvSize: UInt64;    // ǰ֡ۼյ

    FLastFrame: Boolean;       // һ֡
    FOpCode: TWSOpCode;        // 
    FMsgType: TWSMsgType;      // ǰڴϢ

    FMask: TWSMask;            // 
    FMaskBit: PByte;           // ָʾλ
    FMaskExists: Boolean;      // з

    FJSON: TBaseJSON;          // JSONã
    FStream: TMemoryStream;    // ԭʼ/JSON 

    function  CheckInIOCPFlag(ABuf: PAnsiChar; ASize: Integer): TWSMsgType; {$IFDEF USE_INLINE} inline; {$ENDIF}
    function  GetContentSize(InSize: Cardinal): Integer; {$IFDEF USE_INLINE} inline; {$ENDIF}
    function  GetFrameSize(Byte2: Byte): Cardinal; {$IFDEF USE_INLINE} inline; {$ENDIF}

    procedure ExtractFrame(ABuf: PAnsiChar; ASize: Integer; RecvData: Boolean);
    procedure IncRecvCount(RecvCount: Cardinal); {$IFDEF USE_INLINE} inline; {$ENDIF}

    procedure ScanRecvWSBuffers;
  protected
    procedure InitResources(ASize: Cardinal); virtual; abstract;
    procedure InterReceiveData(ASize: Cardinal); virtual; abstract;
    procedure SaveRemainData; virtual; abstract;
    procedure WriteStream(ASize: Cardinal); virtual;abstract;
  public
    constructor Create(AOwner: TObject; AJSON: TBaseJSON);
    destructor Destroy; override;
    procedure Clear; override;
    procedure Prepare(const AData: PAnsiChar; const ASize: Cardinal); override;
  public
    property OpCode: TWSOpCode read FOpCode;
  end;

  // ============ WebSocket ݽ =============

  // չϢ浽
  
  TWSServerReceiver = class(TWSBaseReceiver)
  private
    procedure UnMarkData(ASize: Cardinal);
  protected
    procedure InitResources(ASize: Cardinal); override;
    procedure InterReceiveData(ASize: Cardinal); override;
    procedure SaveRemainData; override;
    procedure WriteStream(ASize: Cardinal); override;
  public
    procedure ClearMark(var Data: PAnsiChar; Overlapped: POverlapped);
    procedure Receive(const AData: PAnsiChar; const ASize: Cardinal); override;
  end;

  // ============ WebSocket ͻݽ =============

  TOnReceiveData   = procedure(Result: TBaseJSON; FrameSize, RecvSize: Int64) of object;

  // ׼ո¼
  TAttachmentEvent = procedure(Result: TBaseJSON) of object;

  // ύӦò¼
  TPostJSONEvent   = procedure(Result: TBaseJSON; OpCode: TWSOpCode;
                               MsgType: TWSMsgType; Stream: TMemoryStream) of object;

  // չϢ

  TWSClientReceiver = class(TWSBaseReceiver)
  private
    FOnAttachment: TAttachmentEvent; // ¼
    FOnPost: TPostJSONEvent;         // ύ¼
    FOnReceive: TOnReceiveData;      // ¼
    procedure PostJSON(AMsgType: TWSMsgType);
  protected
    procedure InitResources(ASize: Cardinal); override;
    procedure InterReceiveData(ASize: Cardinal); override;
    procedure SaveRemainData; override;
    procedure WriteStream(ASize: Cardinal); override;
  public
    destructor Destroy; override;
    procedure Receive(const AData: PAnsiChar; const ASize: Cardinal); override;
  public
    property OnAttachment: TAttachmentEvent read FOnAttachment write FOnAttachment;
    property OnPost: TPostJSONEvent read FOnPost write FOnPost;
    property OnReceive: TOnReceiveData read FOnReceive write FOnReceive;    
  end;

implementation

uses
  iocp_zlib, iocp_utils, iocp_sockets, iocp_clients, iocp_wsClients;

type
  TIOCPDocumentRef = class(TIOCPDocument);
  TResultParamsRef = class(TResultParams);
  TServerWebSocket = class(TWebSocket);

{ TBaseReceiver }

procedure TBaseReceiver.Clear;
begin
  InterInit(False, True);
end;

constructor TBaseReceiver.Create;
begin
  inherited;
  FCheckPassed := True;
  FComplete := True;
end;

procedure TBaseReceiver.IncBufferPos(Offset: Cardinal);
begin
  Inc(FBuffer, Offset);  // ڴַǰ
  Dec(FBufSize, Offset); // δݼ
  Inc(FRecvCount, Offset);  // +
end;

procedure TBaseReceiver.InterInit(ACancel, AComplete: Boolean);
begin
  FCancel := ACancel;
  FCheckPassed := True;
  FComplete := AComplete;
  FErrorCode := 0;
end;

procedure TBaseReceiver.Reset;
begin
  InterInit(True, True);
end;

procedure TBaseReceiver.SetComplete(const Value: Boolean);
begin
  InterInit(False, Value);
end;

{ TCSBaseReceiver }

procedure TCSBaseReceiver.Clear;
begin
  inherited;
  FOwner.Clear;
end;

procedure TCSBaseReceiver.ExtractMessage;
begin
  // ȡһϢ
  //   ReadHead: ҪȡЭϢ
  //       ʽIOCP_SOCKET_FLAG + TMsgHead + [Hash + Hash] + [Data]
  // ϢܴϢĸڴݰ

  // 1. ȡЭϢ
  if FReadHead then  // ȡЭϢ
  begin
    FOwner.SetHeadMsg(PMsgHead(FBuffer + IOCP_SOCKET_FLEN));
    IncBufferPos(IOCP_SOCKET_SIZE);  //  Hash λ
    FReadHead := False;
  end;

  // 1.1 ȡȣǰ
  FComplete := False;  // δ
  FMainLackSize := FOwner.DataSize;
  FAttachLackSize := FOwner.AttachSize;

  if (FBufSize = 0) then  // ֻЭͷ
    IncRecvCount(0, 1)
  else  // 1.2 ȡУ
    case FOwner.CheckType of
      ctMurmurHash:
        GetCheckCodes;
      ctMD5:
        GetCheckCodes;
    end;

  // 2. 
  if (FMainLackSize = 0) then
    ReceiveMainFinish   // 
  else begin
    // ׼ռ
    FOwner.Main.Initialize(FMainLackSize, False);
    if (FBufSize > 0) then  // 
      if (FBufSize > FMainLackSize) then // жϢ˲֣
        WriteMainStream(FMainLackSize)
      else
        WriteMainStream(FBufSize);
  end;

end;

procedure TCSBaseReceiver.IncRecvCount(RecvCount, DataType: Cardinal);
begin
  // յһݣͳ
  IncBufferPos(RecvCount);  // ƽ
  case DataType of // Ƿȱ-
    1: Dec(FMainLackSize, RecvCount);
    2: Dec(FAttachLackSize, RecvCount);
  end;
end;

procedure TCSBaseReceiver.OwnerClear;
begin
  //  FOwner 
  if (FMainLackSize = 0) and (FAttachLackSize = 0) then
    FOwner.Clear;
end;

procedure TCSBaseReceiver.ReceiveAttachmentFinish;
begin
  // ݽ
  // 1. У
  if (FOwner.CheckType > ctNone) then
    VerifyAttachmentStream
  else  // У
    FCheckPassed := True;
  // 2. ѹ
  if FCheckPassed and Assigned(FOwner.Attachment) then
    UZipTreatAttachmentStream;   // 
end;

procedure TCSBaseReceiver.ReceiveMainFinish;
begin
  // ݽϣֻЭͷ
  // 1. У壻 2. ѹ 3. д

  if (FOwner.DataSize = 0) then
    FCheckPassed := True
  else begin
    // 1. У
    if (FOwner.CheckType > ctNone) then
      VerifyMainStream
    else  // У
      FCheckPassed := True;

    // 2. ѹ
    if FCheckPassed then
    begin
      UZipPretreatMainStream;  // ѹԤ
      if (FOwner.VarCount > 0) then  // 
        FOwner.Initialize(FOwner.Main);
    end;
  end;

  // 3. д
  if FCheckPassed and (FOwner.AttachSize > 0) then
  begin
    CreateAttachment;  // ˲Զ
    if (FBufSize > 0) and Assigned(FOwner.Attachment) then // дʣ
      if (FBufSize <= FOwner.AttachSize) then
        WriteAttachmentStream(FBufSize)
      else
        WriteAttachmentStream(FOwner.AttachSize);
  end;

end;

procedure TCSBaseReceiver.RecvSubqueueData;
begin
  // պ

  // 1. д
  if (FMainLackSize > 0) then
    if (FBufSize <= FMainLackSize) then   // 1.1 δ
      WriteMainStream(FBufSize)
    else  // 1.2 ֻ̫д岿
      WriteMainStream(FMainLackSize);

  // 2. д
  if (FBufSize > 0) and (FAttachLackSize > 0) then
    if (FBufSize <= FAttachLackSize) then // 2.1 δ
      WriteAttachmentStream(FBufSize)
    else  // 2.2 ֻ̫д
      WriteAttachmentStream(FAttachLackSize);

  //  FBufSize > 0
end;

procedure TCSBaseReceiver.Reset;
begin
  inherited;
  // ȡý
  FOwner.Cancel;
  FMainLackSize := 0;
  FAttachLackSize := 0;
end;

procedure TCSBaseReceiver.UZipTreatAttachmentStream;
var
  OldFileName, RealFileName: String;
  mStream: TIOCPDocument;
begin
  // ѹ
  
  // 1. ļδѹ -> Ϊԭʧܣ
  // 2. ѹѹļ -> Ϊԭ

  // TIOCPDocument ļرգͷţʹ TIOCPDocument.FileName

  // δȫϣֻ򵥹رգ
  if (FOwner.Action in FILE_CHUNK_ACTIONS) and
     (FOwner.OffsetEnd + 1 < FOwner.Attachment.OriginSize) then
  begin
    FOwner.Attachment.Close; // ֻرգOriginSize 
    Exit;
  end;

  OldFileName := FOwner.Attachment.FileName;
  RealFileName := ExtractFilePath(OldFileName) + FOwner.FileName;

  if (FOwner.ZipLevel = zcNone) then
  begin
    // ֱӹرա
    FOwner.Attachment.Close;
    TIOCPDocumentRef(FOwner.Attachment).RenameDoc(RealFileName);
  end else
  begin
    // Ƚѹļٸ
    mStream := TIOCPDocument.Create(OldFileName + '_UNZIP', True);
    try
      try
        FOwner.Attachment.Position := 0;
        iocp_zlib.ZDecompressStream(FOwner.Attachment, mStream);
      finally
        FOwner.Attachment.Close(True); // رգͬʱɾļ
        TIOCPDocumentRef(mStream).RenameDoc(RealFileName);  // Զر, 
        TIOCPDocumentRef(FOwner.Attachment).FFileName := RealFileName; // ֱӸ
        mStream.Free;  // ͷ
      end;
    except
      FErrorCode := GetLastError;
    end;
  end;

end;

procedure TCSBaseReceiver.UZipPretreatMainStream;
var
  NewBuffers: Pointer;
  NewSize: Integer;
begin
  // ѹԤ
  //   ݱѹ NewBuffersNewBuffers ҵ FOwner.Main 
  if (FOwner.ZipLevel = zcNone) then
    FOwner.Main.Position := 0
  else
    try
      try
        iocp_zlib.ZDecompress(FOwner.Main.Memory, FOwner.DataSize,
                              NewBuffers, NewSize, FOwner.DataSize);
      finally
        FOwner.Main.SetMemory(NewBuffers, NewSize);
      end;
    except
      FErrorCode := GetLastError;
    end;
end;

procedure TCSBaseReceiver.VerifyAttachmentStream;
begin
  // 鸽У
  case FOwner.CheckType of
    ctMurmurHash:  // MurmurHash
      if (FOwner.Action in FILE_CHUNK_ACTIONS) then  // һηΧ Hash
        FCheckPassed := (FCheckCode2.MurmurHash =
                         iocp_mmHash.MurmurHashPart64(FOwner.Attachment.Handle,
                                                      FOwner.Offset, FOwner.AttachSize))
      else  // ļ Hash
        FCheckPassed := (FCheckCode2.MurmurHash =
                         iocp_mmHash.MurmurHash64(FOwner.Attachment.Handle));
    ctMD5:  // MD5
      if (FOwner.Action in FILE_CHUNK_ACTIONS) then // һηΧ MD5
        FCheckPassed := MD5MatchEx(@FCheckCode2.MD5Code,
                                   iocp_md5.MD5Part(FOwner.Attachment.Handle,
                                                    FOwner.Offset, FOwner.AttachSize))
      else  // ļ MD5
        FCheckPassed := MD5MatchEx(@FCheckCode2.MD5Code,
                                   iocp_md5.MD5File(FOwner.Attachment.Handle));
    else
      FCheckPassed := True;
  end;
end;

procedure TCSBaseReceiver.VerifyMainStream;
begin
  // У
  case FOwner.CheckType of
    ctMurmurHash:  // MurmurHash
      FCheckPassed := (FCheckCode.MurmurHash =
                         iocp_mmHash.MurmurHash64(FOwner.Main.Memory, FOwner.DataSize));
    ctMD5:  // MD5
      FCheckPassed := MD5MatchEx(@FCheckCode.MD5Code,
                        iocp_md5.MD5Buffer(FOwner.Main.Memory, FOwner.DataSize));
    else
      FCheckPassed := True;
  end;   
end;

procedure TCSBaseReceiver.WriteAttachmentStream(ByteCount: Cardinal);
begin
  // дݵ
  if Assigned(FOwner.Attachment) then
  begin
    FOwner.Attachment.Write(FBuffer^, ByteCount);
    IncRecvCount(ByteCount, 2); 
    if FComplete then  //  -> У顢ѹ
      ReceiveAttachmentFinish;
  end else
    IncRecvCount(ByteCount, 2);
end;

procedure TCSBaseReceiver.WriteMainStream(ByteCount: Cardinal);
begin
  // дݵ壨д룬ƽ
  FOwner.Main.Write(FBuffer^, ByteCount);
  IncRecvCount(ByteCount, 1);  // ƽ
  if (FMainLackSize = 0) then  // 
    ReceiveMainFinish;
end;

{ TServerReceiver }

constructor TServerReceiver.Create(AOwner: TReceivePack);
begin
  inherited Create;
  FOwner := AOwner;
end;

procedure TServerReceiver.CreateAttachment;
begin
  // Ӧò㽨
end;

procedure TServerReceiver.GetCheckCodes;
begin
  // ȡУ룺
  //   ʽIOCP_HEAD_FLAG + TMsgHead + [У + У] + []
  //     TBaseMessage.GetCheckCode
  case FOwner.CheckType of
    ctMurmurHash: begin  // MurmurHash=64λ
      if (FOwner.DataSize > 0) then   // У
      begin
        FCheckCode.MurmurHash := PMurmurHash(FBuffer)^;
        IncBufferPos(HASH_CODE_SIZE);
      end;
      if (FOwner.AttachSize > 0) then // У
      begin
        FCheckCode2.MurmurHash := PMurmurHash(FBuffer)^;
        IncBufferPos(HASH_CODE_SIZE);
      end;
    end;
    ctMD5: begin  // MD5=128λ
      if (FOwner.DataSize > 0) then   // У
      begin
        FCheckCode.MD5Code := PMD5Digest(FBuffer)^;
        IncBufferPos(HASH_CODE_SIZE * 2);
      end;
      if (FOwner.AttachSize > 0) then // У
      begin
        FCheckCode2.MD5Code := PMD5Digest(FBuffer)^;
        IncBufferPos(HASH_CODE_SIZE * 2);
      end;
    end;
  end;
end;

procedure TServerReceiver.IncRecvCount(RecvCount, DataType: Cardinal);
begin
  inherited;
  //    Ӧò
  case DataType of
    1: FComplete := (FMainLackSize = 0);
    2: FComplete := (FAttachLackSize = 0);
  end;
end;

procedure TServerReceiver.Prepare(const AData: PAnsiChar; const ASize: Cardinal);
begin
  // ݰ׼Դ
  //   ֻһ壬ݲڵǰݰ!
  FBuffer := AData;
  FBufSize := ASize;

  FCancel := False;
  FReadHead := False; // ǰѶȡЭϢ
  IncBufferPos(IOCP_SOCKET_SIZE);  // ݴ

  ExtractMessage;  // ȡϢ
end;

procedure TServerReceiver.Receive(const AData: PAnsiChar; const ASize: Cardinal);
begin
  // պ
  //  ͻ˷͸ʱҪ͸ݲһ
  FBuffer := AData;
  FBufSize := ASize;
  if (ASize = IOCP_CANCEL_LENGTH) and MatchSocketType(FBuffer, IOCP_SOCKET_CANCEL) then
    Reset
  else
    RecvSubqueueData;
end;

{ TClientReceiver }

constructor TClientReceiver.Create(OwnerClass: TReceivePackClass);
begin
  inherited Create;
  FOwnerClass := OwnerClass;   //  TResultParams
  FOwner := OwnerClass.Create; // Ƚһ
end;

procedure TClientReceiver.CreateAttachment;
begin
  // Զ
  TResultParamsRef(FOwner).CreateAttachment(FLocalPath);
end;

destructor TClientReceiver.Destroy;
begin
  if Assigned(FOwner) then
    FOwner.Free;
  inherited;
end;

procedure TClientReceiver.GetCheckCodes;
  procedure WriteMurmurHash(Hash: PIOCPHashCode);
  begin
    if (FBufSize >= HASH_CODE_SIZE) then  // ݹ
    begin
      Hash^.MurmurHash := PMurmurHash(FBuffer)^;
      IncBufferPos(HASH_CODE_SIZE);
    end else
    begin
      // ݲ۶ϣ
      FCodeLackSize := HASH_CODE_SIZE; // ֽ
      FHashCode := @Hash^.MurmurHash;  // дУĵַ
      if (FBufSize > 0) then
        WriteHashCode(FBufSize);
    end;
  end;
  procedure WriteMD5(MD5: PIOCPHashCode);
  begin
    if (FBufSize >= HASH_CODE_SIZE * 2) then // ݹ
    begin
      MD5^.MD5Code := PMD5Digest(FBuffer)^;
      IncBufferPos(HASH_CODE_SIZE * 2);
    end else
    begin
      // ݲ۶ϣ
      FCodeLackSize := HASH_CODE_SIZE * 2;
      FHashCode := @MD5^.MD5Code;
      if (FBufSize > 0) then
        WriteHashCode(FBufSize);
    end;
  end;
begin
  // ȡУ룺
  //   Ϣ͹㲥ϢУ뱻۶
  if (FBufSize > 0) then
    case FOwner.CheckType of
      ctMurmurHash: begin  // MurmurHash=64λ
        if (FOwner.DataSize > 0) then    // У
          WriteMurmurHash(@FCheckCode);
        if (FBufSize > 0) and (FOwner.AttachSize > 0) then  // У
          WriteMurmurHash(@FCheckCode2);
      end;
      ctMD5: begin  // MD5=128λ
        if (FOwner.DataSize > 0) then    // У
          WriteMD5(@FCheckCode);
        if (FBufSize > 0) and (FOwner.AttachSize > 0) then  // У
          WriteMD5(@FCheckCode2);
      end;
    end;
end;

procedure TClientReceiver.IncRecvCount(RecvCount, DataType: Cardinal);
begin
  inherited;
  // ӣ+ȫ -> Ӧò
  FComplete := (FMainLackSize = 0) and (FAttachLackSize = 0);
end;

procedure TClientReceiver.PostMessage;
begin
  // ϢͶ߳
  try
    FOnPost(FOwner);
  finally
    FReadHead := True;  // Ҫ¶Эͷ
    FRecvCount := 0;    // ճ
    FOwner := FOwnerClass.Create; // 
  end;
{  FOwner.Free;  // debug
  FOwner := FOwnerClass.Create; //   }
end;

procedure TClientReceiver.Prepare(const AData: PAnsiChar; const ASize: Cardinal);
begin
  FBuffer := AData;
  FBufSize := ASize;

  FCancel := False;
  FHeadLackSize := 0;  // Эͷδ۶
  FHashCode := nil;    // Уδ۶
  FReadHead := True;   // ȡЭϢ

  ScanRecvBuffers;     // ɨȡϢ
end;

procedure TClientReceiver.Receive(const AData: PAnsiChar; const ASize: Cardinal);
begin
  // պݣ浽
  //   ͬʱ͸ݣҲܰϢ
  FBuffer := AData;
  FBufSize := ASize;

  // 1. Ϣ۶
  //    1.1 Эͷ۶ϣ1.2 У뱻۶

  if (FHeadLackSize > 0) then  // 1.1 Эͷ۶
  begin
    if (FBufSize >= FHeadLackSize) then
    begin
      // ݹ׼ЭͷϢ
      System.Move(AData^, FHeadBuffers[IOCP_SOCKET_SIZE - FHeadLackSize], FHeadLackSize);

      FOwner.SetHeadMsg(PMsgHead(@FHeadBuffers[IOCP_SOCKET_FLEN]));
      IncBufferPos(FHeadLackSize);  //  Hash λ

      FReadHead := False;  // ٴζЭͷ
      FHeadLackSize := 0;

      ScanRecvBuffers;  // ɨ衢ȡϢ
    end else
    begin
      // ݲ
      System.Move(AData^, FHeadBuffers[IOCP_SOCKET_SIZE - FHeadLackSize], FBufSize);
      Dec(FHeadLackSize, FBufSize);  // ȼ
      IncBufferPos(FBufSize); // ƽ -> FBufSize = 0
    end;
  end else
  if Assigned(FHashCode) then  // 1.2 У뱻۶
  begin
    if (FBufSize >= FCodeLackSize) then
      WriteHashCode(FCodeLackSize)
    else
      WriteHashCode(FBufSize);
  end;

  // 2. Эͷ + У봦
  if (FHeadLackSize = 0) and (FHashCode = nil) then
  begin
    // 2.1 պ̻򸽼
    if (FBufSize > 0) then
      RecvSubqueueData;
    // 2.2 ȡϢ
    if (FBufSize > 0) then
      ScanRecvBuffers;
  end;
end;

procedure TClientReceiver.ReceiveAttachmentFinish;
begin
  inherited;
  try
    if (FCheckPassed = False) and Assigned(FOnError) then
      FOnError(FOwner); // У쳣 -> 
  finally
    PostMessage;  // ȫϣͶ
  end;
end;

procedure TClientReceiver.ReceiveMainFinish;
begin
  inherited;
  // ݡͱŵ FOnReceive
  try
    if (FCheckPassed = False) and Assigned(FOnError) then
      FOnError(FOwner)  // У쳣 -> 
    else
    if Assigned(FOnReceive) then  // ý¼
      FOnReceive(FOwner, mdtEntity, FRecvCount, FAttachLackSize = 0);
  finally
    if (FCheckPassed = False) or (FOwner.AttachSize = 0) then
      PostMessage;  // ϣͶ
  end;
end;

procedure TClientReceiver.ScanRecvBuffers;
begin
  // ݰ׼Դ
  //   ʽIOCP_SOCKET_FLAG + TMsgHead + [Hash + Hash] + [Data]
  //   1. ͻ˿жϢģҪֽ⣡
  //   2. ϢʱЭͷܱ۶

  // ȡϢ
  while (FReadHead = False) and (FBufSize > 0) or (FBufSize >= IOCP_SOCKET_SIZE) do
    ExtractMessage; 

  // ʣ࣬ӦС IOCP_SOCKET_SIZE
  if (FBufSize > 0) then
  begin
    // ʣ೤ < Эͷ, ۶ϣ
    // ʣݵ FHeadBuffers´ƴ
    FComplete := False;
    FHeadLackSize := IOCP_SOCKET_SIZE - FBufSize;  // ȱֽ
    System.Move(FBuffer^, FHeadBuffers[0], FBufSize);
    IncBufferPos(FBufSize);
  end;
end;

procedure TClientReceiver.WriteAttachmentStream(ByteCount: Cardinal);
begin
  inherited; // ByteCountǰյֽ
  if Assigned(FOnReceive) then  // ý¼
    FOnReceive(FOwner, mdtAttachment, FRecvCount, FAttachLackSize = 0);
end;

procedure TClientReceiver.WriteHashCode(ByteCount: Cardinal);
begin
  // У뱻۶ϣȱյĲ
  System.Move(FBuffer^, FHashCode^, ByteCount);
  IncBufferPos(ByteCount);     // ƽ
  Inc(FHashCode, ByteCount);   // ƽ´дλ
  Dec(FCodeLackSize, ByteCount);  // ֽ-
  if (FCodeLackSize = 0) then  // дУ
    FHashCode := nil;
end;

{ TWSBaseReceiver }

function TWSBaseReceiver.CheckInIOCPFlag(ABuf: PAnsiChar; ASize: Integer): TWSMsgType;
begin
  // չ INIOCP_JSON_FLAG
  if (ABuf = nil) then
    Result := mtDefault
  else
  if (ASize >= INIOCP_JSON_FLAG_LEN) and
     (PInIOCPJSONField(ABuf)^ = INIOCP_JSON_FLAG) then
    Result := mtJSON
  else
    Result := mtDefault;
end;

procedure TWSBaseReceiver.Clear;
begin
  // յϢ 
  FData := nil;
  FFrameSize := 0;
  FFrameRecvSize := 0;
  if Assigned(FStream) and (FStream.Size > 0) then
    FStream.Clear;
  inherited;
end;

constructor TWSBaseReceiver.Create(AOwner: TObject; AJSON: TBaseJSON);
begin
  inherited Create;
  FOwner := AOwner;
  FJSON := AJSON;
end;

destructor TWSBaseReceiver.Destroy;
begin
  Clear;
  if Assigned(FStream) then
    FStream.Free;
  inherited;
end;

procedure TWSBaseReceiver.ExtractFrame(ABuf: PAnsiChar; ASize: Integer; RecvData: Boolean);
var
  i: Integer;
  iByte: Byte;
  OffSet: Integer;

  procedure _MovePos(Step: Integer);
  begin
    Inc(ABuf, Step);
    Dec(ASize, Step);  //  < 0
    Inc(OffSet, Step);
  end;

begin
  // ֡ṹȡϢ
  // ֡Ȳ ASize >= 2

  OffSet := 0;
  FFrameRecvSize := 0;  // ǰ֡=0

  // 1.  1 ֽ

  // ǷΪĩ֡
  iByte := PByte(ABuf)^;
  FLastFrame := (iByte shr 7 > 0);

  // λ RSV1RSV2RSV3

  // ȡ
  iByte := (iByte and $0F);
  if (iByte in WEBSOCKET_OPCODES) then
    FOpCode := TWSOpCode(iByte)
  else
    FOpCode := ocClose;  // 쳣رգ

  // 2.  2 ֽ
  _MovePos(1);

  // λ
  FMaskBit := PByte(ABuf);
  iByte := PByte(ABuf)^;

  // Ƿ룺λͻ
  FMaskExists := (iByte shr 7 > 0);

  // ֡ȣ7λ
  FFrameSize := (iByte and $7F);

  // 3. ȡ֡
  _MovePos(1);     //  3 ֽ

  case FFrameSize of
    126: begin     // <= max(UInt16)ǰ2ֽ -> 
      FFrameSize := 0;
      if (ASize >= 2) then
      begin
        TByteAry(@FFrameSize)[1] := TByteAry(ABuf)[0];
        TByteAry(@FFrameSize)[0] := TByteAry(ABuf)[1];
      end;
      _MovePos(2); // 
    end;
    127: begin     // <= max(UInt64)ǰ8ֽ -> 
      FFrameSize := 0;
      if (ASize >= 8) then
        for i := 0 to 7 do
          TByteAry(@FFrameSize)[7 - i] := TByteAry(ABuf)[i];
      _MovePos(8); // 
    end;
  end;

  // 4. ȡ루4ֽڣͻޣ
  if FMaskExists then  // 
  begin
    if (ASize >= 4) then
      FMask := PWSMask(ABuf)^;
    _MovePos(4);   // 
  end;

  // 5. 
  //    ASize = 0 -> ս
  //    ASize > 0 -> 
  if RecvData and (ASize >= 0) then
  begin
    IncBufferPos(OffSet);  // ƶ FBuffer
    InterReceiveData(ASize);  // 
  end;

end;

function TWSBaseReceiver.GetContentSize(InSize: Cardinal): Integer;
begin
  // жҪдĳ
  if (InSize = 0) then
    Result := 0
  else
  if (FFrameRecvSize + InSize <= FFrameSize) then
    Result := InSize
  else
    Result := FFrameSize - FFrameRecvSize;
end;

function TWSBaseReceiver.GetFrameSize(Byte2: Byte): Cardinal;
begin
  // ֡ĳ
  //   ͻˣClose/Ping/Pong  2 ֽڣСֵ
  case (Byte2 and $7F) of
    126: Result := 4;
    127: Result := 10;
    else Result := 2;
  end;
end;

procedure TWSBaseReceiver.IncRecvCount(RecvCount: Cardinal);
begin
  // ճ+
  Inc(FFrameRecvSize, RecvCount);
  // Ƿ
  FComplete := FLastFrame and (FFrameRecvSize = FFrameSize);
end;

procedure TWSBaseReceiver.Prepare(const AData: PAnsiChar; const ASize: Cardinal);
begin
  FBuffer := AData;
  FBufSize := ASize;
  ScanRecvWSBuffers;  // ɨ
end;

procedure TWSBaseReceiver.ScanRecvWSBuffers;
begin
  // ݰ
  //   1. տʱжϢҪֽ⣡
  //   2. ϢʱЭͷܱ۶

  if Assigned(FHeadAddr) then   // ƴȡ֡Ϣ
  begin
    ExtractFrame(@FHeader[0], FHeadAddr - PAnsiChar(@FHeader[0]), False);  // ݸչ
    InterReceiveData(FBufSize); // ǰ֡ݣԶȡ GetContentSize(FBufSize)
    FHeadAddr := nil;
  end;
     
  while (FBufSize >= 2) and (FBufSize >= GetFrameSize(Byte((FBuffer + 1)^))) do
    ExtractFrame(FBuffer, FBufSize, True);

  if (FBufSize > 0) then  // һϢ
    SaveRemainData;
end;

{ TWSServerReceiver }

procedure TWSServerReceiver.ClearMark(var Data: PAnsiChar; Overlapped: POverlapped);
var
  i: Integer;
  Buf: PAnsiChar;
begin
  // ޸ϢԱ㷢ظͻ˻㲥
  //  ָʾ룬ǰ
  if FMaskExists and Assigned(FMaskBit) then
  begin
    FData := Data; // Data  TWSData.Data
    Buf := PAnsiChar(FData - 4);
    for i := 1 to FFrameRecvSize do
    begin
      Buf^ := FData^;
      Inc(Buf); Inc(FData);
    end;

    // ־
    FMaskBit^ := FMaskBit^ and Byte($7F);
    FMaskBit := nil;

    Dec(Overlapped^.InternalHigh, 4);
    Dec(Data, 4);    
  end;
end;

procedure TWSServerReceiver.InitResources(ASize: Cardinal);
begin
  // ʼԴ
  //  JSON ־ Socket 

  FData := FBuffer;  // ݿʼλ
  if (FOpCode = ocText) then  //  JSON ־TBaseJSON.SaveToStream
    FMsgType := CheckInIOCPFlag(FBuffer, ASize);

  case FMsgType of
    mtDefault:      // 
      TServerWebSocket(FOwner).SetProps(FOpCode, mtDefault, Pointer(FData),
                                        FFrameSize, ASize);

    mtJSON: begin   // JSON Ϣ
      if (Assigned(FStream) = False) then
        FStream := TMemoryStream.Create;
      FStream.Size := FFrameSize; // к֡ʱдʱԶ
      TServerWebSocket(FOwner).SetProps(FOpCode, mtJSON,
                                        Pointer(FData), 0, 0);
    end;
    
    mtAttachment:
      TServerWebSocket(FOwner).SetProps(FOpCode, mtAttachment,
                                        Pointer(FData), 0, 0);
  end;
end;

procedure TWSServerReceiver.InterReceiveData(ASize: Cardinal);
var
  RecvCount: Cardinal;
begin
  // ׼
  RecvCount := GetContentSize(ASize);  // ȡЧ
  try
    //  -> 
    if (RecvCount > 0) then
      UnMarkData(RecvCount);
    // ׼Դ
    InitResources(RecvCount);
    // 浽
    if (RecvCount > 0) then
      WriteStream(RecvCount);
  finally
    // ƽ
    if (RecvCount > 0) then
      IncBufferPos(RecvCount);
  end;
end;

procedure TWSServerReceiver.Receive(const AData: PAnsiChar; const ASize: Cardinal);
var
  RecvCount: Integer;
begin
  // պ

  FBuffer := AData;
  FBufSize := ASize;
  FData := FBuffer;  // ݿʼλ

  // 1. ǰ֡Ľ
  if (FFrameRecvSize < FFrameSize) then  // FLackSize > 0 ʱ  FFrameSize = 0
  begin
    RecvCount := GetContentSize(FBufSize);
    UnMarkData(RecvCount);   // 루ˣ
    WriteStream(RecvCount);  // 浽
    IncBufferPos(RecvCount); // ǰ
  end;

  // 2. ʣһϢ
  if (FBufSize > 0) then
    SaveRemainData;
    
end;

procedure TWSServerReceiver.SaveRemainData;
begin
  // ʣݣ´νգ
  // ͻ˷ʱ΢ͣ٣ʣ
end;

procedure TWSServerReceiver.UnMarkData(ASize: Cardinal);
var
  i: Integer;
  p: PByte;
begin
  //  -> xor 
  if (ASize > 0) and FMaskExists then
  begin
    p := PByte(FBuffer);
    for i := FFrameRecvSize to FFrameRecvSize + ASize - 1 do
    begin
      p^ := p^ xor FMask[i mod 4];
      Inc(p);
    end;
  end;
end;

procedure TWSServerReceiver.WriteStream(ASize: Cardinal);
begin
  // ݵ

  // ӽ
  IncRecvCount(ASize);

  // 1. д
  case FMsgType of
    mtDefault:     //  mtDefault Ϣ
      Inc(TServerWebSocket(FOwner).FMsgSize, ASize);
    mtJSON:        // չ JSON Ϣ
      FStream.Write(FBuffer^, ASize);
    mtAttachment:  // 
      if Assigned(FJSON.Attachment) then
        FJSON.Attachment.Write(FBuffer^, ASize);
  end;

  // 2. 
  if FComplete then
    case FMsgType of
      mtJSON: begin  // JSON Ϣ
        FJSON.Initialize(FStream, True);  // ת JSON, ͬʱ FStream
        if FJSON.HasAttachment then // 
          FMsgType := mtAttachment  // ´Ϊ
        else
          FMsgType := mtDefault;
      end;
      mtAttachment:  // 
        FMsgType := mtDefault;  // ´Ϊ mtDefault
    end;
    
end;     

{ TWSClientReceiver }

destructor TWSClientReceiver.Destroy;
begin
  if Assigned(FJSON) then
    FJSON.Free;
  inherited;
end;

procedure TWSClientReceiver.InitResources(ASize: Cardinal);
begin
  // ׼
  //   JSONչݾ浽
  if (Assigned(FStream) = False) then
    FStream := TMemoryStream.Create;
end;

procedure TWSClientReceiver.InterReceiveData(ASize: Cardinal);
var
  RecvCount: Cardinal;          
begin
  // ׼
  RecvCount := GetContentSize(ASize);  // ȡЧ
  try
    // ׼Դ
    InitResources(RecvCount);
    // 浽
    WriteStream(RecvCount);
  finally
    // ƽ
    if (RecvCount > 0) then
      IncBufferPos(RecvCount);
  end;
end;

procedure TWSClientReceiver.PostJSON(AMsgType: TWSMsgType);
begin
  // Ͷ JSON Ϣ
  try
    if (AMsgType = mtDefault) then  // Ͷ
      FOnPost(FJSON, FOpCode, AMsgType, FStream)
    else
      FOnPost(FJSON, FOpCode, AMsgType, nil);
  finally
    if (AMsgType = mtDefault) then
      FStream := TMemoryStream.Create;
    FJSON := TJSONResult.Create(FOwner);
  end;
end;

procedure TWSClientReceiver.Receive(const AData: PAnsiChar; const ASize: Cardinal);
var
  RecvCount: Integer;
begin
  // պ

  FBuffer := AData;
  FBufSize := ASize;

  // 1. ǰ֡Ľ
  if (FFrameRecvSize < FFrameSize) then  // FLackSize > 0 ʱ  FFrameSize = 0
  begin
    RecvCount := GetContentSize(FBufSize);
    WriteStream(RecvCount);  // 浽
    IncBufferPos(RecvCount); // ǰ
  end;

  // 2. ʣ

  // 2.1 ֡۶
  // ⣺ͻ˽յϢʱ֡ᱻ۶Ϸֿ

  if (FBufSize > 0) and (FLackSize > 0) then
  begin
    if (FLackSize = 9) then  // ת壬֡ĵڶֽ
    begin
      FHeader[1] := FBuffer^;
      FLackSize := GetFrameSize(Byte(FHeader[1])) - 2;  // ȡ
      IncBufferPos(1);
    end;

    //  ocCloseocPingocPong -> RecvCount = 0
    if (FBufSize < FLackSize) then
      RecvCount := FBufSize
    else
      RecvCount := FLackSize;

    if (RecvCount > 0) then
    begin
      System.Move(FBuffer^, FHeadAddr^, RecvCount);
      IncBufferPos(RecvCount);
      Inc(FHeadAddr, RecvCount);
      Dec(FLackSize, RecvCount);
    end;
  end;

  // 3. ֡ȡ
  if (FLackSize = 0) then
    ScanRecvWSBuffers;

end;

procedure TWSClientReceiver.SaveRemainData;
var
  i: Integer;
begin
  //  2 ֽڣocCloseocPingocPong
  // ʱݵ FHeader´ƴ

  if (FBufSize = 1) then
  begin
    FHeader[0] := FBuffer^;
    FHeadAddr := @FHeader[2];  // 3ֽλã TWSClientReceiver.Receive
    FLackSize := 9;  // ת壬 ٵڶֽ
  end else
  begin
    for i := 0 to FBufSize - 1 do
      FHeader[i] := (FBuffer + i)^;
    // ĳ
    FLackSize := GetFrameSize(Byte(FHeader[1])) - FBufSize;
    if (FLackSize > 0) then
      FHeadAddr := @FHeader[FBufSize]
    else begin  // ֡ȡϢ
      ExtractFrame(PAnsiChar(@FHeader[0]), FBufSize, False);
      FComplete := FFrameSize = 0;  // ocCloseocPing ocPong
    end;
  end;

  if (FLackSize > 0) then
  begin
    FFrameSize := 0;
    FFrameRecvSize := 0;
    FComplete := False;
  end;

  FBufSize := 0;
    
end;

procedure TWSClientReceiver.WriteStream(ASize: Cardinal);
begin
  // 1. ӽ
  IncRecvCount(ASize);

  // 2. 浽FMsgTypeδ֪
  if (ASize > 0) then
    if (FOpCode = ocText) then  // ıϢδ֪InIOCP-JSON
      FStream.Write(FBuffer^, ASize)
    else
    if Assigned(FJSON.Attachment) then  // 
    begin
      FJSON.Attachment.Write(FBuffer^, ASize);
      FOnReceive(FJSON, FFrameSize, ASize);  // Ľս
    end;

  // 5. 
  if FComplete then
  begin
    // 
    if (FOpCode = ocBiary) then
      FMsgType := mtAttachment
    else
      FMsgType := CheckInIOCPFlag(PAnsiChar(FStream.Memory), FStream.Size);

    // 
    FFrameSize := 0;
    FFrameRecvSize := 0;
    
    case FMsgType of
      mtDefault:
        if (FOpCode <= ocClose) then
          PostJSON(mtDefault);  // ͬʱͶ FStream

      mtJSON: begin
        FJSON.Initialize(FStream, True);  // ת JSON, ͬʱ FStream
        if FJSON.HasAttachment then  // ͬ
        begin
          FMsgType := mtAttachment;  // ´Ϊ
          if Assigned(FOnAttachment) then
            FOnAttachment(FJSON); // жǷļ
        end else  // ֱͶ
        begin
          FMsgType := mtDefault;  // ´Ϊ mtDefault
          PostJSON(mtJSON);
        end;
      end;

      mtAttachment: begin
        FMsgType := mtDefault;   // ´Ϊ mtDefault
        if Assigned(FJSON.Attachment) then
          FJSON.Attachment.Position := 0;
        PostJSON(mtAttachment);  // ͶŸ
      end;
    end;

  end;

end;

end.

